@@ -5,27 +5,38 @@ def unify(term1, term2):
 
     # https://www.javatpoint.com/ai-unification-in-first-order-logic 
 
+    # quick check - if equal then return
+    if (term1 == term2): return {}
+
     # Step. 1: If Ψ1 or Ψ2 is a variable or constant, then:
     if (type(term1) is str or type(term2) is str):
     
         sTerm1 = str(term1)
 
         #  if term1 is a standalone variable
-        if (sTerm1.startswith("?")) and (" " in sTerm1 == False):
+        if (sTerm1.startswith("?")) and (" " not in sTerm1):
         
             #  then just unify it with term2
             unification = {}
-            unification.Add(sTerm1, term2)
+            unification[sTerm1] = term2
             return unification
         
         else:
         
             sTerm2 = str(term2)
 
-            # return Text.TextbookUnifyStrings(sTerm1, sTerm2)
-            return unify_strings(sTerm1, sTerm2)
+            #  if term2 is a standalone variable
+            if (sTerm2.startswith("?")) and (" " not in sTerm2):
         
-    
+                #  then just unify it with term1
+                unification = {}
+                unification[sTerm2] = term1
+                return unification
+            
+            else:
+                # return Text.TextbookUnifyStrings(sTerm1, sTerm2)
+                return unify_strings(term1, sTerm2)
+         
     elif (type(term1) is dict and type(term2) is dict):
 
         joTerm1 = term1
@@ -44,14 +55,24 @@ def unify(term1, term2):
 
         # Step. 5: For i = 1 to the number of elements in Ψ1.
         for prop in joTerm2.keys():
-        
+    
+            # ignore #seq (positional) information, as it is injected by the engine at runtime
+            if (prop == "#seq-start"): continue
+            if (prop == "#seq-end"): continue
+
+            # can still use #seq-start and #seq-end in rules, just need to escape them with an extra #
+            prop1 = prop
+            if (prop == "##seq-start"): prop1 = "#seq-start"
+            if (prop == "##seq-end"): prop1 = "#seq-end"
+
             # a) Call Unify function with the ith element of Ψ1 and ith element of Ψ2, and put the result into S.
-            if prop not in joTerm1: return None
-            jtTest1 = joTerm1[prop]
+            if prop1 not in joTerm1: return None
+            jtTest1 = joTerm1[prop1]
 
             # b) If S = failure then returns Failure
             subSet2 = unify(jtTest1, joTerm2[prop])
             if (subSet2 is None): return None
+            
             # c) If S ≠ NIL then do,
             elif (len(subSet2.keys()) > 0): subSet = {**subSet, **subSet2}
         
@@ -190,8 +211,19 @@ def apply_unification(term, unification):
     if (type(term) is dict):
         result = {}
         for key in term.keys():
-            newval = apply_unification(term[key], unification)
-            result[key] = newval
+            if (key == "#combine"):
+                items_to_combine = term["#combine"]
+                newval = {}
+                for item in items_to_combine:
+                    new_item = apply_unification(item, unification)
+                    newval = {**new_item, **newval}
+                # assume combine is a standalone operation
+                # could also merge this will other keys
+                # result[key] = newval
+                return newval
+            else:
+                newval = apply_unification(term[key], unification)
+                result[key] = newval
         return result
 
     elif (type(term) is list):
@@ -203,9 +235,16 @@ def apply_unification(term, unification):
 
     elif (type(term) is str):
         # substitute unification into the then part
-        for key in unification.keys(): 
-            term = term.replace(key, unification[key])
-        return term
+        if str(term).startswith("#\\"):
+            pass
+        elif (term.startswith("?")) and (" " not in term):
+            if term in unification:
+                return unification[term]
+        else:
+            for key in unification.keys():
+                new_val = str(unification[key])
+                term = term.replace(key, new_val)
+            return term
 
     else:
         return term
